import time

from mineshaft import Mineshaft
from mineshaft import needs_resolved

try:
    from graphite_api.intervals import Interval, IntervalSet
    from graphite_api.node import LeafNode, BranchNode
except ImportError:
    from graphite.intervals import Interval, IntervalSet  # noqa
    from graphite.node import LeafNode, BranchNode  # noqa


class MineshaftLeafNode(LeafNode):
    __fetch_multi__ = 'mineshaft'


class MineshaftBranchNode(BranchNode):
    pass


class MineshaftFinder(object):
    __fetch_multi__ = 'mineshaft'
    __slots__ = ('driver',)

    def __init__(self, config=None):
        if config is not None:
            url = config.get('mineshaft', {}).get('url')
        else:
            from django.conf import settings
            url = getattr(settings, 'MINESHAFT_URL')
        self.driver = Mineshaft(url)

    def find_nodes(self, query):
        if needs_resolved(query.pattern):
            search = query.pattern
            if query.pattern == '*':
                search = ''
            elif query.pattern[-2:] == '.*':
                search = query.pattern[:-2]
            if needs_resolved(search):
                nodes = self.driver.resolve(query.pattern)
            else:
                nodes = self.driver.children(search)
            for node in nodes:
                if node.leaf:
                    yield MineshaftLeafNode(node.key, MineshaftReader(node.key, self))
                else:
                    yield MineshaftBranchNode(node.key)
        else:
            yield MineshaftLeafNode(query.pattern, MineshaftReader(query.pattern, self))

    def fetch_multi(self, nodes, start_time, end_time):
        step = end_time - start_time
        series = {}
        paths = [node.path for node in nodes]
        data = self.driver.metrics(paths, start_time, end_time)
        if data:
            # Collect all the series into their own dict
            series = {key: value['series'] for key, value in data.iteritems()}

            # Extract the time info out of the first key
            src = data.keys()[0]
            start_time = data[src]['from']
            end_time = data[src]['to']
            step = data[src]['step']

        return (start_time, end_time, step), series


class MineshaftReader(object):
    __slots__ = ('path', 'finder')

    def __init__(self, path, finder):
        self.path = path
        self.finder = finder

    def fetch(self, start_time, end_time):
        data = self.finder.driver.metrics(self.path, start_time, end_time).get(self.path, {})
        if not data:
            return (start_time, end_time, end_time - start_time), []
        return (data['from'], data['to'], data['step']), data['series']

    def get_intervals(self):
        # FIXME: mineshaft server should allow for obtaining this data
        start = time.time() - 3600 * 2
        end = max(start, time.time())
        return IntervalSet([Interval(start, end)])



"""
mineshaft
~~~~~~~~~

:copyright: (c) 2014 by Matt Robenolt.
:license: BSD, see LICENSE for more details.
"""
from time import time
from operator import attrgetter
import requests

__all__ = ('Mineshaft', 'Node', 'LeafNode', 'BranchNode')


getkey = attrgetter('key')
HINTS = {'*', '[', '{', '?'}


def needs_resolved(path):
    """Test if a path is a pattern or is already fully qualified"""
    for c in path:
        if c in HINTS:
            return True
    return False


def is_leaf(node):
    return node.leaf


class Node(object):
    depth = 0
    leaf = False
    key = None

    def __repr__(self):
        return '<%s key=%r depth=%r>' % (type(self).__name__, self.key, self.depth)

    def __str__(self):
        return str(self.key)

    def __unicode__(self):
        return unicode(self.key)


def load(raw):
    if raw['Leaf']:
        node = LeafNode()
    else:
        node = BranchNode()
    node.depth = raw['Depth']
    node.key = raw['Key']
    return node


class LeafNode(Node):
    leaf = True


class BranchNode(Node):
    leaf = False


class Mineshaft(object):
    """Library for interacting with the mineshaft API"""

    def __init__(self, address='http://localhost:8080'):
        self.address = address.rstrip('/')
        self.session = requests.Session()

    def get(self, url, **kwargs):
        return self.session.get(self.address+url, **kwargs).json()

    def children(self, query=''):
        return self.to_nodes(self.get('/children', params={'query': query}))

    def resolve(self, query):
        return self.to_nodes(self.get('/paths', params={'query': query}))

    def metrics(self, targets, start_time=None, end_time=None):
        if start_time is None:
            start_time = int(time()) - 86400  # 24 hours ago
        if end_time is None:
            end_time = int(time())

        assert start_time > 0
        assert end_time > 0
        assert start_time < end_time

        if isinstance(targets, basestring):
            targets = [targets]

        targets = set(targets)
        unresolved = set(filter(needs_resolved, targets))
        resolved = targets - unresolved
        if unresolved:
            targets = resolved | set(map(getkey, filter(is_leaf, self.resolve(unresolved))))

        params = {'target': targets, 'from': start_time, 'to': end_time}
        return self.get('/metrics', params=params)

    def intervals(self, targets):
        return []

    def ping(self):
        try:
            r = self.get('/ping')
            return r['status'] == 200
        except Exception:
            return False

    def to_nodes(self, results):
        return [load(n) for n in results]



#!/usr/bin/env python
"""
graphite-mineshaft
~~~~~~~~~~~~~~~~~~

:copyright: (c) 2014 by Matt Robenolt.
:license: BSD, see LICENSE for more details.
"""
from setuptools import setup

install_requires = ['mineshaft']


with open('README.rst') as f:
    long_description = f.read()


setup(
    name='graphite-mineshaft',
    version='0.0.1',
    author='Matt Robenolt',
    author_email='matt@ydekproductions.com',
    url='https://github.com/mattrobenolt/graphite-mineshaft',
    description='Graphite backend for Mineshaft',
    license='BSD',
    long_description=long_description,
    py_modules=['graphite_mineshaft.finders'],
    install_requires=install_requires,
    zip_safe=True,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python',
        'Topic :: Software Development',
    ],
)
